import yaml

try:
    import ujson as json
except ImportError:
    import json

from os import path
from abc import ABC

from .filecache import FileCache


class FileDriver(ABC):
    def __init__(self, filepath: str):
        self.filepath = filepath
        filename, extension = path.splitext(self.filepath)
        if not extension:
            if path.exists(filename) is False:
                self.filepath += '.' + type(self).__name__.lower()

    def handle(self):
        raise NotImplementedError

    @classmethod
    def read(cls, filepath: str) -> dict:
        cache = FileCache(filepath)
        if cache.get():
            return cache.get()

        obj = cls(filepath)
        cache.set(obj.handle())
        return cache.get()


class Yaml(FileDriver):
    def handle(self) -> dict:
        with open(self.filepath, 'r') as f:
            return yaml.load(f)


class Json(FileDriver):
    def handle(self) -> dict:
        with open(self.filepath, 'r') as f:
            return json.load(f)



class Singleton(type):
    _instances = {}

    def __call__(cls, filepath):
        if not filepath in cls._instances:
            cls._instances[filepath] = super(Singleton, cls).__call__(filepath)

        return cls._instances[filepath]


class FileCache(metaclass=Singleton):
    def __init__(self, filepath):
        self.filepath = filepath
        self._data = None

    def get(self):
        return self._data

    def set(self, data):
        self._data = data



from abc import ABC
from os import environ

from .drivers import Yaml, Json


class BaseLoader(ABC):
    @classmethod
    def isdisabled(cls, disabled):
        return cls.__name__ in disabled

    @classmethod
    def isenabled(cls, enabled):
        if not enabled:
            return True
        return cls.__name__ in enabled

    @classmethod
    def read(cls):
        raise NotImplementedError


class Environment(BaseLoader):
    """ Usage:
            Environment.read(prefix, source)

        Args:
            "prefix" will be added to all variables,
            for example: prefix_variable in uppercase

            "source" is a dictionary of names for finding variables,
            result will contain only variables that will be found.
    """

    @classmethod
    def read(cls, *args):
        result = {}
        for k, v in args[1].items():
            var_name = (args[0] + '_' + k).upper()
            value = environ.get(var_name, None)
            if value:
                result.update({k: value})
        return result


class File(BaseLoader):
    """ Usage:
            File.read(prefix, source, file_path)

        It works in the same way as the environment class,
        except for the additional argument containing
        path to the readable file.
    """

    drivers = (
        Json,
        Yaml
    )

    @classmethod
    def read(cls, *args):
        for driver in cls.drivers:
            try:
                result = {}
                read = driver.read(args[2])[args[0]]
                for k, v in args[1].items():
                    value = read.get(k, None)
                    if value:
                        result.update({k: value})
                return result
            except Exception:
                continue
        return {}



__version__ = '0.1.4'

from abc import ABC
from copy import deepcopy
from os import environ, getcwd

from .loaders import File, Environment


class ABConfig(ABC):
    loaders = (
        File,
        Environment
    )

    enabled_loaders = ()
    disabled_loaders = ()

    filepath = environ.get('CONFIG_PATH', getcwd() + '/config')

    def __str__(self):
        return type(self).__name__

    def __getitem__(self, item: str):
        return self._data[item]

    def __getattr__(self, attr):
        return getattr(self._data, attr)

    def __init__(self):
        self._data = self.make(*self._attrs)

    def make(self, *args):
        result = {}
        for name in args:
            defaults = getattr(self, name)
            result_for_name = deepcopy(defaults)
            for loader in self.loaders:
                if (loader.isdisabled(self.disabled_loaders) is True and
                        loader.isenabled(self.enabled_loaders) is False):
                    continue
                result_for_name.update(loader.read(name, defaults, self.filepath))
            result.update({name: result_for_name})
        return result

    def get(self, item=None):
        if not item:
            return self._data
        elif item not in self._data:
            raise KeyError(f'Variable "{item}" was not exist in "{self}" config')
        return self._data[item]

    @property
    def _attrs(self):
        result = []
        excluded_attrs = (
            'enabled_loaders',
            'disabled_loaders',
            'filepath',
        )
        for attr in type(self).__dict__.keys():
            if (attr[:2] != '__' and attr[:1] != '_' and
                    attr not in excluded_attrs):
                result.append(str(attr))
        return result
